#
# 2203.15556.R, 28 Apr 24
#
# Data from:
# Training Compute-Optimal Large Language Models
# Jordan Hoffmann, Sebastian Borgeaud, Arthur Mensch, Elena Buchatskaya, Trevor Cai, Eliza Rutherford,
# Diego de Las Casas, Lisa Anne Hendricks, Johannes Welbl, Aidan Clark, Tom Hennigan, Eric Noland,
# Katie Millican, George van den Driessche, Bogdan Damoc, Aurelia Guy, Simon Osindero, Karen Simonyan,
# Erich Elsen, Jack W. Rae, Oriol Vinyals and Laurent Sifre
# Figure 4.

library("colorspace")

par(bty="l")
par(las=1)
par(pch="+")
pal_col=rainbow(2)

# Extracted from pdf
x_vert_dashed=c(83.85625, # y-axis
		138.481407,
		154.054886,
		187.548148,
		208.680043,
		224.253523,
		257.746784,
		278.878679,
		294.452159,
		327.94542)

FLOPS=data.frame(x=x_vert_dashed[c(1, 3, 6, 9)],
			budget=c(1e18, 1e19, 1e20, 1e21))

x_base=FLOPS$x[1] # measuring up from 10^18
x_expand=FLOPS$x[2]-FLOPS$x[1] # X length representing a factor of 10

# Map internal pdf x/y to visible axis values
model_size=data.frame(y=c(374.286119, 413.527645),
			size=c(40e9, 100e9))

y_base=model_size$y[2] # Measuring down from 100B
y_expand=model_size$y[2]-model_size$y[1] # Y length representing a factor of 10/4 -> 2.5

# extracted from 2203-pdf_all.txt using pdf_fig-4.awk
pt_info=read.csv("pdf_fig-4.csv", as.is=TRUE)

pt_info$x_val=cumsum(pt_info$x) # pdf positions are relative to the previous
pt_info$y_val=cumsum(pt_info$y) # pdf positions are relative to the previous

# Handle log scales
pt_info$FLOPS=1e18*10^((pt_info$x_val-x_base)/x_expand)
pt_info$size=100e9/(2.5^((y_base-pt_info$y_val)/y_expand))

write.csv(pt_info, file="size-FLOP.csv", row.names=FALSE)


plot(pt_info$FLOPS, pt_info$size, log="xy",
	xlab="Training FLOPS", ylab="Model size")

pt_mod=glm(log(size) ~ log(FLOPS), data=pt_info)
summary(pt_mod)


# Calculate consistent svg y-axis mappings
# Values from replication extraction code
y_mapping=function(y_size)
{
return (26.872+(66.113-26.872)*(log(100)-log(y_size))/(log(100)-log(40)))
}


y_mapping(c(10, 1, 0.1))

