# -*- coding: utf-8 -*-
"""approach_3_svg.ipynb

Automatically generated by Colab.

Original file is located at
    https://colab.research.google.com/drive/1ROmEyJHuXrVc3Jr62iLeWLSzjemjl8mC

#Setup
"""

import matplotlib.pyplot as plt
from google.colab import files
from bs4 import BeautifulSoup
import pandas as pd
import numpy as np
import re
import io
import base64
from io import BytesIO
from PIL import Image

class LogScale:
    """Converts SVG coordinates into data coordinates"""

    def __init__(self, svg_range, data_range):
        self.svg_range = np.array(svg_range)
        self.data_range = np.array(data_range)

    def get_coord(self, svg_coord):
        alpha = (svg_coord - self.svg_range[0])/(self.svg_range[1] - self.svg_range[0])
        data_coord = self.data_range[0] * (self.data_range[1]/self.data_range[0])**alpha
        return data_coord

import requests

# Define the URL of the SVG file hosted on GitHub
url = "https://raw.githubusercontent.com/Besiroglu/data/main/approach_3.svg"

# Send a GET request to the URL to fetch the SVG file content
response = requests.get(url)

# Ensure the request was successful
if response.status_code == 200:
    svg_content = response.text
    print("SVG Content:")
    print(svg_content)
else:
    print(f"Failed to fetch SVG file. Status code: {response.status_code}")

# # Upload the SVG file from your local machine.
# uploaded = files.upload()

# # The `uploaded` variable is a dictionary where the keys are the filenames
# # and the values are the uploaded data.
# # Let's assume you've uploaded a single SVG file, we'll read its contents next.

# filename = next(iter(uploaded))
# print(f"Uploaded file: {filename}")

# # Now, read the content of the SVG file as text.
# with open(filename, 'r') as file:
#     svg_content = file.read()

# print("SVG Content:")
# print(svg_content)

"""# Extract position data from SVG

"""

# Parse the SVG data with BeautifulSoup
soup = BeautifulSoup(svg_content, 'xml')

# Get the group of points on the left plot
left_scatter_group = soup.find('g', {'id': 'left-points'})

# Get all points in the group
left_scatter_points = left_scatter_group.find_all('g', {'clip-path': True})

# Create lists to store the extracted data
x_values = []
y_values = []
colors = []

# Extract location and color for each scatter point
for point in left_scatter_points:
    # Check if the <g> element contains exactly two <path> elements
    paths = point.find_all('path')
    if len(paths) == 2:
        # Extract the fill color from the first <path>
        color = paths[0].get('fill')

        # Extract the position from the transform attribute of the first <path>
        transform = paths[0].get('transform')
        if transform:
            # Use regular expression to extract x and y values
            match = re.search(r'matrix\(1,0,0,-1,(\d+\.?\d*),(\d+\.?\d*)\)', transform)
            if match:
                x = float(match.group(1))
                y = float(match.group(2))

                # Append the extracted data to the respective lists
                x_values.append(x)
                y_values.append(y)
                colors.append(color)

# Create a DataFrame with the extracted data
df_left = pd.DataFrame({'x': x_values, 'y': y_values, 'color': colors})
df_left.head()

# Convert from SVG coordinates to data coordinates

# Y scale
y_tick_svg_coords = [26.872, 66.113, 124.290, 221.707, 319.125]
y_tick_data_coords = [100e9, 40e9, 10e9, 1e9, 100e6]
y_scale = LogScale([y_tick_svg_coords[0], y_tick_svg_coords[-1]], [y_tick_data_coords[0], y_tick_data_coords[-1]])

# Check scale
print('y_scale:')
for svg_coord, data_coord in zip(y_tick_svg_coords, y_tick_data_coords):
    print(f"  {svg_coord}px -> {y_scale.get_coord(svg_coord):.2e} (expected: {data_coord:.2e})")
print()

# X scale
x_tick_svg_coords = [83.856, 154.055, 224.254, 294.452]
x_tick_data_coords = [1e18, 1e19, 1e20, 1e21]
x_scale = LogScale([x_tick_svg_coords[0], x_tick_svg_coords[-1]], [x_tick_data_coords[0], x_tick_data_coords[-1]])

# Check scale
print('x_scale:')
for svg_coord, data_coord in zip(x_tick_svg_coords, x_tick_data_coords):
    print(f"  {svg_coord}px -> {x_scale.get_coord(svg_coord):.2e} (expected: {data_coord:.2e})")
print()

df_left['model size'] = df_left['y'].apply(y_scale.get_coord)
df_left['compute_budget'] = df_left['x'].apply(x_scale.get_coord)

df_left
df_left.head()

plt.figure(figsize=(8, 6))
for i, row in df_left.iterrows():
      plt.scatter(row['compute_budget'], row['model size'], color=row['color'])
plt.xlabel('compute_budget')
plt.ylabel('Model size')
plt.yscale('log')
plt.xscale('log')
plt.title('')
plt.grid(True)
plt.gca().set_axisbelow(True)
plt.show()

"""# Color scheme to loss mapping"""

# Extract the base64 data from the provided string
loss_bar = soup.find('image', {'id': 'image_308'})
base64_data = loss_bar.get('xlink:href').split(',')[1]

# Decode the base64 data to binary
binary_data = base64.b64decode(base64_data)

# Create a BytesIO object from the binary data
img_bytes = BytesIO(binary_data)

# Load the image using Pillow
img = Image.open(img_bytes)

# Display the image
img

# Loss scale
tick_svg_coords = [26.872, 101.343, 197.353, 332.671]
tick_data_coords = [5, 4, 3, 2]
loss_scale = LogScale([tick_svg_coords[0], tick_svg_coords[-1]], [tick_data_coords[0], tick_data_coords[-1]])

# Check scale
print('loss_scale:')
for svg_coord, data_coord in zip(tick_svg_coords, tick_data_coords):
    print(f"  {svg_coord}px -> {loss_scale.get_coord(svg_coord):.2f} (expected: {data_coord:.2f})")

# Position of the top of the image
match = re.search(r'matrix\(1,0,0,1,(\d+\.?\d*),(\d+\.?\d*)\)', loss_bar.parent.get('transform'))
img_y = float(match.group(2))

# Choose a specific width position (x value), here we take the middle of the image
x = img.width // 2

# Initialize a list to hold the hex values and their corresponding height
hex_values = []

# Loop through each pixel along the height at this x position
for y in range(img.height):
    # Get the RGB value of the pixel
    pixel = img.getpixel((x, y))
    # Convert RGB to hex, slicing to skip the alpha value if present
    hex_value = '#{:02x}{:02x}{:02x}'.format(*pixel[:3])
    hex_values.append([y, hex_value, loss_scale.get_coord(y + img_y)])

# Create a DataFrame from the hex values with a column for the pixel height and hex color
df_color = pd.DataFrame(hex_values, columns=['pixel_height', 'hex_color', 'loss_value'])

# Display the DataFrame
df_color # Show the first few rows of the dataframe to check

# Group by hex_color and find pairs with the same hex_color
groups = df_color.groupby('hex_color')

# Initialize variables to track the maximum difference and the corresponding pair of points
max_diff = 0
pair_with_max_diff = None

for color, group in groups:
    # If there's at least two points with the same color, calculate differences
    if len(group) > 1:
        for i in range(len(group) - 1):
            for j in range(i + 1, len(group)):
                diff = abs(group.iloc[i]['loss_value'] - group.iloc[j]['loss_value'])
                if diff > max_diff:
                    max_diff = diff
                    pair_with_max_diff = (group.iloc[i], group.iloc[j])

pair_with_max_diff, max_diff

# Group df_color by 'hex_color' and calculate the median 'loss_value' for each color
color_loss_median = df_color.groupby('hex_color')['loss_value'].median().reset_index()

# Merge the median loss values with df_left based on color
df_left_merged = pd.merge(df_left, color_loss_median, how='left', left_on='color', right_on='hex_color')
df_left_merged

"""# Download"""

df_left_merged

df_left_merged.rename(columns={
    'compute_budget': 'Training FLOP',
    'loss_value': 'loss',
    'model size': 'Model Size'
}, inplace=True)

# Import necessary libraries
from google.colab import files
import pandas as pd

# Assuming df_left_merged is your DataFrame
# Save the DataFrame to a CSV file
df_left_merged.to_csv('df_left_merged.csv', index=False)

# Download the file to your local system
files.download('df_left_merged.csv')

